package Main;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import diffeq.DiffEqSolver;
import diffeq.DynamicModel;

// This is the main class of the project. It instantiates a model and a
// differential equation solver and then it uses the solver to step through
// time in small increments and it tabulates simulation results in a file,
// suitable for further processing.
// This is the class, implementing the dynamic model of the Rossler-like system
// with the opamps. All three integrators and the non-linearity are programmed
// explicitly and can easily be recognized in the source code.
// In this class only static variables are used (not really OO-paradigm), but
// as we only need one instance this is not a major issue.
public class BromateHydroxylamine {

    static final public String outFile = "/home/woelen/reaction.txt";
    
    public static void main(String[] args) {

        // Create instance of the Rossler-like model. This model contains the
        // equations and it knows its own order (number of independent states).
        DynamicModel reactionModel = new ReactionModel();


        // Instantiate a differential equation solver and allocate space
        // for storing the state variables of the model. The number of states
        // is queried from the model itself.
        DiffEqSolver solver = new DiffEqSolver(reactionModel);
        double[] y = new double[reactionModel.getOrder()];

        BufferedWriter stream = null;
        try {
            int progress = 0;

            // Open the output file to which the simulation results are written
            // in tabular form. Post-processing can be done with e.g. Excel,
            // Open Office or Gnuplot.
            File f = new File(outFile);
            stream = new BufferedWriter(new FileWriter(f));

            // Set the initial condition for all state variables.
            y[0] = ReactionModel.InitBrO3;    // Initial concentration of BrO3(-)
            y[1] = ReactionModel.InitBr;      // Initial concentration of Br(-)
            
            // Initial concentration of H(+) ion is determined by the (weak) acidity
            // of the hydroxylammonium ion. The constant K2 is nothing more than 1/Kz
            // for the weak acid ion NH3OH(+): K2 = 1/Kz = [NH3OH(+)]/([H(+)]*[NH2OH]).
            // From this it is possible to estimate the initial concentration of H(+).
            y[2] = Math.sqrt(ReactionModel.InitNH3OH / ReactionModel.K2);
            
            y[3] = ReactionModel.InitNH3OH;    // Initial concentration of NH3OH(+)
            y[4] = ReactionModel.InitBr2;      // Initial concentration of Br2

            // Perform a simulation in steps of 0.01 second, for a period
            // of 100 seconds. Parameters h, relacc and absacc are parameters
            // for the differential equation solver. 
            double dt = 0.01;
            double tmax = 100.0;
            double h = 0.001;
            double relacc = 1e-7;
            double absacc = 1e-9;
            // Write all states to the output file.
            for (double t = 0.0; t <= tmax; t += dt) {
                stream.write("" + t);
                for (int i = 0; i < y.length; i++) {
                    stream.write("  " + y[i]);
                }
                solver.solve(y, t, y, t + dt, h, absacc, relacc);
                stream.newLine();

                // For each 2500 steps print a single dot. This gives feedback
                // to the user that the program is still working.
                progress++;
                if (progress % 2500 == 0) {
                    System.out.print(".");
                    System.out.flush();
                }
            }

            // Print message that the program is finished.
            System.out.println(" DONE");
        } 
        catch (Exception ex) {
            System.err.println(ex);
        } 
        finally {
            if (stream != null) {
                try {
                    stream.close();
                } 
                catch (IOException ignore) {
                }
            }
        }
    }
}




class ReactionModel implements DynamicModel {

    // Initial concentrations of chemicals in the solutions.
    static final double InitBrO3 = 1.0;
    static final double InitBr = 0.0;
    static final double InitNH3OH = 4.0;
    static final double InitBr2 = 0.0;
    
    // K1 and K2 simply are 1, divided by the Ka values of the
    // involved acids. These acid equilibria add two algebraic
    // constraints, which lead to a somewhat more complicated
    // rate law for reaction of bromate ion with hydroxylammonium.
    static final public double K1 = Math.pow(10, 0.70);    // 1/Ka of HBrO3
    static final public double K2 = Math.pow(10, 5.98);    // 1/Ka of NH3OH(+)
    
    // Rate constants for reaction of bromate with bromide
    // BrO3(-) + Br(-) + 2H(+) --> HOBr + HBrO2
    // r1 = (kBr*[Br(-)] + kBrS*[Br(-)]*[Br(-)])*[BrO3(-)]*[H(+)]*[H(+)]
    static final public double kBr = 2.025;
    static final public double kBrS = 0.7;
    
    // Rate constant for reaction of bromate with hydroxylammonium ion
    // BrO3(-) + NH3OH(+) + H(+) --> HBrO2 + [NH2(OH)2](+)
    static final public double kN = 2.57e3;
    
    // Rate constant for reaction of bromine with hydroxylammonium ion.
    // The net reaction is the one, given below, the reaction is
    // first order in [Br2] and in [NH3OH(+)], but the mechanism is
    // quite complicated. This is not relevant for this model.
    // Br2 + NH3OH(+) --> 2Br(-) + 3H(+) + ½H2O + ½N2O
    static final public double kBr2 = 1.7e3;

    // The reaction between HOBr and HBrO2, bromide ion and H(+) is
    // VERY fast, compared to the other reactions, described above.
    // The reaction from the species [NH2(OH)2](+) to N2O, H2O and
    // H(+) also is extremely fast, compared to the reactions,
    // described above. A good approximation does not need state
    // variables for these reactions.
    
    
    // The set of differential equations, used by the model of the
    // reaction between bromate ion and hydroxylamine. Input is the
    // time and the state variables at the given time. The derivatives
    // of the state variables are computed and stored in the output
    // array xdot[].
    @Override
    public void computeDerivatives(double[] xdot, double t, double[] x) {
        double BrO3 = x[0];    // Concentration of BrO3(-)
        double Br = x[1];      // Concentration of Br(-)
        double H = x[2];       // Concentration of H(+)
        double NH3OH = x[3];   // Concentration of NH3OH(+)
        double Br2 = x[4];     // Concentration of Br2

        // All rate constants are multiplied by some heat factor. This is a very
        // coarse model, but it takes into account that the reaction is exothermic
        // and that while the reaction runs, the liquid heats up. This hardly has
        // any effect, until the near blow-up occurs.
        // For the heat factor, the difference between the initial amount of
        // bromate and the actual amount of bromate is used. From experiment it
        // is known that the liquid heats up close to boiling hot. We also assume
        // roughly double reaction speed at each 10 degrees increase of temperature.
        // The concentration of bromate is in the order of magnitude of 1 mol/l
        // and if all is gone, we assume 70 degrees increase of temperature, which
        // leads to a 2^7 factor for the rate constants. 
        // It is assumed that no loss of heat occurs. At the very small timescale
        // in which the explosively fast reaction occurs, this is valid.
        double heatFactor = Math.pow(2.0, 7*(ReactionModel.InitBrO3 - BrO3));
        
        // Determine the rate factors, according to the papers, mentioned in the
        // webpage. The heat factor is added to the equations to take into account
        // the heating of the liquid during the reaction.
        double r1 = heatFactor * (kBr + kBrS * Br) * Br * BrO3 * H * H;
        double r2 = heatFactor * kN * K1 * H * NH3OH * BrO3 / ((1+K1*H) * K2);
        double r3 = heatFactor * kBr2 * Br2 * NH3OH;

        //xdot[0] = -r1 - r2;
        //xdot[1] = -5*r1 - 3*r2 + 2*r3;
        //xdot[2] = -6*r1 - 3*r2 + 3*r3;
        //xdot[3] = -r2 - r3;
        //xdot[4] = 3*r1 + 2*r2 - r3;
        
        double a = x[1]/(x[1]+x[3]);
        double b = x[3]/(x[1]+x[3]);
        
        xdot[0] = -r1 -r2;
        xdot[1] = -r1 - a*4*r1 + b*2*r1 - a*3*r2 + b*r2 + 2*r3;
        xdot[2] = -2*r1 - a*4*r1 + b*5*r1 - a*3*r2 + b*3*r2 + 3*r3;
        xdot[3] = -b*3*r1 - r2 - b*2*r2 - r3;
        xdot[4] = -r3 + a*3*r1 + a*2*r2;
    }

    
    
    // The order of this set of equations is fixed and equal to 5.
    @Override
    public int getOrder() {
        return 5;
    }
}
