package diffeq;

// Class for Runge-Kutta solution of sets of ordinary differential equations.
public class DiffEqSolver {

    private static double SAFETY = 0.9D; // safety scaling factor for Runge
                                         // Kutta Fehlberg tolerance check
    private double[] k1;
    private double[] k2;
    private double[] k3;
    private double[] k4;
    private double[] k5;
    private double[] k6;
    private double[] y5;
    private double[] y6;
    private double[] yd;
    private double[] ys;
    private double[] dydt;
    private DynamicModel dynamicModel;

    // Constructor, which queries the order from the supplied model
    // and then allocates storage for many intermediate results.
    public DiffEqSolver(DynamicModel model) {
        int n = model.getOrder();
        this.dynamicModel = model;
        k1 = new double[n];
        k2 = new double[n];
        k3 = new double[n];
        k4 = new double[n];
        k5 = new double[n];
        k6 = new double[n];
        y5 = new double[n];
        y6 = new double[n];
        yd = new double[n];
        ys = new double[n];
        dydt = new double[n];
    }

    
    // Runge-Kutta-Cash-Karp for n ordinary differential equations.
    // The output yn may be the same vector as y0. In the latter case
    // the input values are overwritten by the output values.
    private void cashKarp(double[] yn, double t0, double[] y0,
            double tn, double h, double abstol, double reltol, int maxiter) {
        int n = dynamicModel.getOrder();

        // Initialise the starting values for x and y. These variables
        // are adjusted for each integration step.
        double t = t0;
        for (int i = 0; i < n; i++) {
            ys[i] = y0[i];
        }

        // Integrate in small steps, until we reach the end time tn.
        int nIter = 0;
        while (t < tn) {
            nIter++;
            if (nIter > maxiter) {
                throw new ArithmeticException("Maximum number of integration steps exceeded!");
            }

            dynamicModel.computeDerivatives(dydt, t, ys);
            for (int i = 0; i < n; i++) {
                k1[i] = h * dydt[i];
            }

            for (int i = 0; i < n; i++) {
                yd[i] = ys[i] + k1[i] / 5.0;
            }
            dynamicModel.computeDerivatives(dydt, t + h / 5.0, yd);
            for (int i = 0; i < n; i++) {
                k2[i] = h * dydt[i];
            }

            for (int i = 0; i < n; i++) {
                yd[i] = ys[i] + (3.0 * k1[i] + 9.0 * k2[i]) / 40.0;
            }
            dynamicModel.computeDerivatives(dydt, t + 3.0 * h / 10.0, yd);
            for (int i = 0; i < n; i++) {
                k3[i] = h * dydt[i];
            }

            for (int i = 0; i < n; i++) {
                yd[i] = ys[i] + (3.0 * k1[i] - 9.0 * k2[i] + 12.0 * k3[i]) / 10.0;
            }
            dynamicModel.computeDerivatives(dydt, t + 3.0 * h / 5.0, yd);
            for (int i = 0; i < n; i++) {
                k4[i] = h * dydt[i];
            }

            for (int i = 0; i < n; i++) {
                yd[i] = ys[i] - 11.0 * k1[i] / 54.0 + 5.0 * k2[i] / 2.0 - 70.0 * k3[i] / 27.0 + 35.0 * k4[i] / 27.0;
            }
            dynamicModel.computeDerivatives(dydt, t + h, yd);
            for (int i = 0; i < n; i++) {
                k5[i] = h * dydt[i];
            }

            for (int i = 0; i < n; i++) {
                yd[i] = ys[i]
                        + 1631.0 * k1[i] / 55296.0
                        + 175.0 * k2[i] / 512.0
                        + 575.0 * k3[i] / 13824.0
                        + 44275.0 * k4[i] / 110592.0
                        + 253.0 * k5[i] / 4096.0;
            }
            dynamicModel.computeDerivatives(dydt, t + 7.0 * h / 8.0, yd);
            for (int i = 0; i < n; i++) {
                k6[i] = h * dydt[i];
            }

            double maxerr = 0.0;
            for (int i = 0; i < n; i++) {
                y5[i] = ys[i]
                        + 2825.0 * k1[i] / 27648.0
                        + 18575.0 * k3[i] / 48384.0
                        + 13525.0 * k4[i] / 55296.0
                        + 277.0 * k5[i] / 14336.0
                        + k6[i] / 4.0;
                y6[i] = ys[i]
                        + 37 * k1[i] / 378.0
                        + 250.0 * k3[i] / 621.0
                        + 125.0 * k4[i] / 594.0
                        + 512.0 * k6[i] / 1771.0;
                double err = Math.abs(y6[i] - y5[i]);
                double tol = Math.abs(y5[i]) * reltol + abstol;
                maxerr = Math.max(maxerr, err / tol);
            }
            if (maxerr <= 1.0) {
                // Accept the solution and advance to the next x and y values.
                t += h;
                for (int i = 0; i < n; i++) {
                    ys[i] = y5[i];
                }

                // Compute next stepsize, but never increase stepsize
                // by more than a factor of 4 and never advance beyond
                // the end value xn.
                double delta = SAFETY * Math.pow(maxerr, -0.2);
                delta = (delta <= 4.0) ? delta : 4.0;
                h *= delta;
                if (t + h > tn) {
                    h = tn - t;
                }
            } 
            else {
                // Do not advance to the next x and y values. So, x and
                // yn are not modified.

                // Compute next stepsize, but never decrease by more
                // than a factor of 10.
                double delta = SAFETY * Math.pow(maxerr, -0.25);
                h *= (delta < 0.1) ? 0.1 : delta;
            }
        }

        for (int i = 0; i < n; i++) {
            yn[i] = ys[i];
        }
    }

    
    
    // This method solves the set of differential equations at time tn,
    // where initial conditions are given at time t0. The arguments must
    // be used 
    /**
     * Solves the set of differential equations, which are passed to the
     * constructor of the DiffEqSolver object. Initial conditions at time t0 are
     * provided and this method computes the state variables at time tn.
     *
     * @param yn This is an output parameter. It is filled with the value of the
     * state variables at time tn. The array yn may be the same as the array y0.
     * If that is the case, then the initial condition will be overwritten with
     * the final result.
     * @param t0 The time for which the initial conditions are given.
     * @param y0 The initial state at time t0.
     * @param tn The end time for which the state must be computed.
     * @param h The initial stepsize used for integrating from t0 to tn. This
     * stepsize can (and usually will) be adjusted by the internal software of
     * the differential equation solved. The value provided as input is an
     * initial suitable guess.
     * @param abstol
     * @param reltol The last two parameters are stop criteria for the
     * integration routine. When the error estimate goes below |y|*reltol +
     * abstol then the integration stops and the computed value is accepted and
     * copied to the output prarameter yn.
     */
    public void solve(double[] yn, double t0, double[] y0,
            double tn, double h, double abstol, double reltol) {

        double nsteps = (tn - t0) / h;
        int maxiter = (int) (nsteps * 100);

        if (h > tn - t0) {
            h = tn - t0;
        }

        cashKarp(yn, t0, y0, tn, h, abstol, reltol, maxiter);
    }
}
